; win64 driver written in FASM for usage with fdbg
; implemented functions:
; - start driver	start + DispatchCreateClose procedures
; - stop driver		DispatchCreateClose + DriverUnload procedures
; - write to a driver	DispatchWrite procedure


number_of_interrupts	=	100h		; we have int00-intFF
number_of_exceptions	=	20h		; only int00-int1F are exceptions
size_of_gate_entry	=	16		; each gate in 64-bit descriptor table is 16 bytes
size_of_address		=	8		; we are in 64-bit world, 64 bits = 8 bytes


format PE64 native 5.02 at 10000h on 'hdr.exe'
entry start


include 'fdbg.inc'
include 'KMD64.inc'
include	'fdbg_version.inc'
include	'msrAMD64.inc'
include	'macro_align.inc'


section '.text' code readable executable notpageable

start:
; lpDriverObject, lpusRegistryPath
; rcx=pDriverObject rdx=pDriverPath

	push	rbx
	sub	rsp,8*(4+8)		; reserve 12 qwords on the stack, 8 for us
					; Low 4 qwords may be destroyed by API, we can't use them
; RSP is now aligned 16

	mov	rbx,rcx			; Save RCX to RBX. RBX is protected against destroing by API.

; set DebugCtlMSR according driver settings flags
	mov	ecx,DebugCtlMSR
	rdmsr
	mov	[DebugCtlMSR_eax_old],eax
	mov	[DebugCtlMSR_edx_old],edx
	and	al,not 11b
	or	al,[DebugCtlMSR_settings]
	wrmsr

; load structure to point to IRP handlers
virtual at 0
DriverObject	DRIVER_OBJECT
end virtual

	lea	rdx,[DriverUnload]
	lea	rcx,[DispatchCreateClose]
	lea	rax,[DispatchWrite]
	mov	[rbx + DriverObject.DriverUnload],rdx
	mov	[rbx + DriverObject.MajorFunction + IRP_MJ_CREATE_OFFSET],rcx
	mov	[rbx + DriverObject.MajorFunction + IRP_MJ_CLOSE_OFFSET],rcx
	mov	[rbx + DriverObject.MajorFunction + IRP_MJ_WRITE_OFFSET],rax

	lea	rdx,[cusDevice_string]
	lea	rcx,[rsp+8*(4+4)]	; use 2 quadwords reserved on the stack
	call	qword [RtlInitUnicodeString]

; create and initialize device object
	lea	rax,[rsp+8*(4+3)]	; 1 quadword for [DeviceObject] - we use stack for it
	xor	edx,edx			; zeroing rdx, 2nd param for API
	mov	[rsp+8*(4+2)],rax	; pass address for [DeviceObject] - 7th arg to the API
	mov	byte [rsp+8*(4+1)],dl	; false -> rdx = 0
	mov	[rsp+8*(4+0)],edx	; 5th arg for API
	mov	r9d,FILE_DEVICE_UNKNOWN	; 4th arg for API
	lea	r8,[rsp+8*(4+4)]	; [rsp+8*(4+4)] created by RtlInitUnicodeString
	mov	rcx,rbx			; lpDriverObject - 1st arg for API
	call	qword [IoCreateDevice]

if STATUS_SUCCESS = 0
	or	eax,eax			; 2 bytes opcode - smaller than cmp eax,0
else
	cmp	eax,STATUS_SUCCESS
end if
	jnz	driver_return		; break on error

	; create symbolic link to the user-visible name
	lea	rdx,[cusSymbolicLink_string]
	lea	rcx,[rsp+8*(4+6)]	; use 2 quadwords reserved on the stack
	call	qword [RtlInitUnicodeString]

	lea	rdx,[rsp+8*(4+4)]	; 2nd arg for API
	lea	rcx,[rsp+8*(4+6)]	; 1st arg for API
	call	qword [IoCreateSymbolicLink]

if STATUS_SUCCESS = 0
	or	eax,eax
else
	cmp	eax,STATUS_SUCCESS
end if
	jz	driver_return

; save status
	mov	ebx,eax

; unsuccess, delete device object
	mov	rcx,[rsp+8*(4+3)]	; 1 quadword for [DeviceObject] - we use stack for it
					; qword is filled when call IoCreateDevice
	call	qword [IoDeleteDevice]

; assign result
	mov	eax,ebx

driver_return:
	add	rsp,8*(4+8)
	pop	rbx
	ret


align 10h
DriverUnload:
;proc DriverUnload lpDriverObject
; rcx = lpDriverObject
	push	rbx
	sub	rsp,8*(4+2)			; reserve 6 qwords on the stack, top 2 qword can be used, low 4 qwords may be destroyed because API can use them.
	mov	rbx,rcx

	call	UnhookExceptions

; restore original DebugCtlMSR
	mov	ecx,DebugCtlMSR
	mov	eax,[DebugCtlMSR_eax_old]
	mov	edx,[DebugCtlMSR_edx_old]
	wrmsr

	lea	rdx,[cusSymbolicLink_string]
	lea	rcx,[rsp+8*(4+0)]		; requires 2 quadwords reserved
	call	qword [RtlInitUnicodeString]
	lea	rcx,[rsp+8*(4+0)]
	call	qword [IoDeleteSymbolicLink]
	mov	rcx,[rbx + DriverObject.DeviceObject]
	call	qword [IoDeleteDevice]
	add	rsp,8*(4+2)
	pop	rbx
	ret


align 10h
DispatchCreateClose:
;proc DispatchCreateClose pDeviceObject, lpIrp
; rcx = pDeviceObject, rdx = lpIrp
	sub	rsp,8*(4+1)			; align 16 stack

virtual at 0
vIRP	IRP
end virtual

	mov	rcx,rdx		; 1st arg for API

if IO_NO_INCREMENT=0
	xor	edx,edx		; 2nd arg for API

 if STATUS_SUCCESS = 0
	mov	[rcx + vIRP.IoStatus.Status],edx
	mov	[rcx + vIRP.IoStatus.Information],rdx
 else
	mov	[rcx + vIRP.IoStatus.Status],STATUS_SUCCESS
	xor	eax,eax
	mov	[rcx + vIRP.IoStatus.Information],rax
 end if

else
	mov	edx,IO_NO_INCREMENT
end if

	call	qword [IofCompleteRequest]

if STATUS_SUCCESS=0
	xor	eax,eax
else
	mov	eax,STATUS_SUCCESS
end if

	add	rsp,8*(4+1)
	ret


align 10h
DispatchWrite:
; function called when writing to the device
;proc DispatchWrite lpDeviceObject, lpIrp
; rcx = lpDeviceObject, rdx = lpIrp
	push	rbx rsi rdi
	sub	rsp,8*(4+0)		; stack is now aligned 16

	mov	ebx,STATUS_UNSUCCESSFUL	; rbx = current status

	mov	rdi,rdx			; rdi = lpIrp = PIRP

	xor	eax,eax
	mov	rsi,[rdi + vIRP.Tail.Overlay.CurrentStackLocation]
	mov	[rdi + vIRP.IoStatus.Information],rax

; word [rsi] must contain IRP_MJ_WRITE = 4
; But we don't check this because we are certainly on DispatchWrite procedure.
; Usefull for branch if one procedure handles more functions, e.g. Write and Read from driver.

; structure used by driver to comunicate with write_device.exe
virtual at 0
fdbgdq	fdbgDriverQuery
end virtual

virtual at 0
iosl	IO_STACK_LOCATION
end virtual

	cmp	[rsi + iosl.Parameters.Write.Length],size_of_fdbgDriverQuery
	jnz	egzyduz			; does number of bytes written to device match or differ ?

; get address of DriverQuery
	mov	rsi,[rdi + vIRP.UserBuffer]	; PDriverQuery

; assume we have failed until we don't have success

if (STATUS_NOT_IMPLEMENTED - STATUS_UNSUCCESSFUL) = 1
	inc	ebx
else
	mov	ebx,STATUS_NOT_IMPLEMENTED
end if

; get iocode from DriverQuery
	mov	eax,[rsi + fdbgdq.iocode]	; eax = user I/O code
	sub	eax,DRIVER_QUERY_START
	jz	DispatchWrite_proc00
	dec	eax
	jz	DispatchWrite_proc01
	dec	eax
	jz	DispatchWrite_proc02
	dec	eax
	jz	DispatchWrite_proc03
	dec	eax
	jz	DispatchWrite_proc04
	dec	eax
	jz	DispatchWrite_proc05
	dec	eax
	jz	DispatchWrite_proc06
	dec	eax
	jz	DispatchWrite_proc07
	dec	eax
	jz	DispatchWrite_proc08
	dec	eax
	jz	DispatchWrite_proc09
	dec	eax
	jnz	egzyduz

DispatchWrite_proc0A:
;; Enabled branches recording? Not relevant, it is disabled after triggering int01...
;; Hooked exceptions? Relevant!
;;	cmp	[exceptions_hooked],0
;; we can use that AL=0 now
;	cmp	[exceptions_hooked],al
;	jz	egzyduz_STATUS_SUCCESS
	lea	rdx,[LastBranchFromIPLo]
	mov	rcx,qword [rdx + LastBranchFromIPLo - LastBranchFromIPLo]
	mov	rax,qword [rdx + LastBranchToIPLo - LastBranchFromIPLo]
	mov	[rsi + fdbgdq.LastBranchFromIP],rcx
	mov	[rsi + fdbgdq.LastBranchToIP],rax
	mov	rcx,qword [rdx + LastExceptionFromIPLo - LastBranchFromIPLo]
	mov	rax,qword [rdx + LastExceptionToIPLo - LastBranchFromIPLo]
	mov	[rsi + fdbgdq.LastExceptionFromIP],rcx
	mov	[rsi + fdbgdq.LastExceptionToIP],rax

; no, we can't be sure whether the structure of driver query is properly aligned
; we must use the above safe and bigger code
;	movdqa	xmm0,dqword [LastBranchFromIPLo]
;	movdqa	xmm1,dqword [LastExceptionFromIPLo]
;	movdqa	dqword [rsi + fdbgdq.LastBranchFromIP],xmm0
;	movdqa	dqword [rsi + fdbgdq.LastExceptionFromIP],xmm1

;align 16
egzyduz_STATUS_SUCCESS:
if STATUS_SUCCESS = 0
	xor	ebx,ebx
else
	mov	ebx,STATUS_SUCCESS
end if

egzyduz:

if IO_NO_INCREMENT=0
	xor	edx,edx
else
	mov	edx,IO_NO_INCREMENT
end if

	mov	[rdi + vIRP.IoStatus.Status],ebx

	mov	rcx,rdi			; lpIrp
	call	qword [IofCompleteRequest]

	mov	eax,ebx			; put result into eax e.g. STATUS_SUCCESS

	add	rsp,8*(4+0)
	pop	rdi rsi rbx
	ret

align 16
DispatchWrite_proc00:
; get version info
	mov	[rsi + fdbgdq.version],fdbg_version_binary
	jmp	egzyduz_STATUS_SUCCESS

align 16
DispatchWrite_proc01:
	movzx	ecx,[DebugCtlMSR_settings]
	movzx	eax,[exceptions_hooked]
	mov	[rsi + fdbgdq.flags0],ecx
	mov	[rsi + fdbgdq.flags1],eax
	jmp	egzyduz_STATUS_SUCCESS

align 16
DispatchWrite_proc02:
; read DebugCtlMSR
	mov	ecx,DebugCtlMSR
	rdmsr
	mov	[rsi + fdbgdq.DebugCtlMSRHi],edx
	mov	[rsi + fdbgdq.DebugCtlMSRLo],eax
	jmp	egzyduz_STATUS_SUCCESS

align 16
DispatchWrite_proc03:
; write DebugCtlMSR
	mov	edx,[rsi + fdbgdq.DebugCtlMSRHi]
	mov	eax,[rsi + fdbgdq.DebugCtlMSRLo]
	mov	ecx,DebugCtlMSR
	wrmsr
	jmp	egzyduz_STATUS_SUCCESS

align 16
DispatchWrite_proc04:
; enable recording branches by setting DebugCtlMSR.LBR
; The processor automatically disables control-transfer recording when a debug exception (int01) occurs by clearing DebugCtlMSR.LBR to 0. The contents of the control-transfer recording MSRs are not altered by the processor when the int01 occurs. Before exiting the debug-exception handler, software can set DebugCtlMSR.LBR to 1 to re-enable the recording mechanism.
	mov	ecx,DebugCtlMSR
	rdmsr
	or	al,1
	wrmsr
	or	[DebugCtlMSR_settings],1
	jmp	egzyduz_STATUS_SUCCESS

align 16
DispatchWrite_proc05:
; disable recording branches by setting DebugCtlMSR.LBR
; The processor automatically disables control-transfer recording when a debug exception (int01) occurs by clearing DebugCtlMSR.LBR to 0. The contents of the control-transfer recording MSRs are not altered by the processor when the int01 occurs. Before exiting the debug-exception handler, software can set DebugCtlMSR.LBR to 1 to re-enable the recording mechanism.
	mov	ecx,DebugCtlMSR
	rdmsr
	and	al,not 1
	wrmsr
	and	[DebugCtlMSR_settings],not 1
	jmp	egzyduz_STATUS_SUCCESS


align 16
DispatchWrite_proc06:
; enable single step on branches by setting DebugCtlMSR.BTF
; The processor automatically disables control-transfer single stepping when a debug exception (int01) occurs by clearing DebugCtlMSR.BTF to 0. rFLAGS.TF is also cleared when an int01 exception occurs. Before exiting the debug-exception handler, software must set both DebugCtlMSR.BTF and rFLAGS.TF to 1 to restart single stepping.
; When enabled, this single-step mechanism causes an int01 exception to occur on every branch instruction, interrupt, or exception.
	mov	ecx,DebugCtlMSR
	rdmsr
	or	al,10b
	wrmsr
	or	[DebugCtlMSR_settings],10b
	jmp	egzyduz_STATUS_SUCCESS

align 16
DispatchWrite_proc07:
; enable single step on branches by setting DebugCtlMSR.BTF
; The processor automatically disables control-transfer single stepping when a debug exception (int01) occurs by clearing DebugCtlMSR.BTF to 0. rFLAGS.TF is also cleared when an int01 exception occurs. Before exiting the debug-exception handler, software must set both DebugCtlMSR.BTF and rFLAGS.TF to 1 to restart single stepping.
; When enabled, this single-step mechanism causes an int01 exception to occur on every branch instruction, interrupt, or exception.
	mov	ecx,DebugCtlMSR
	rdmsr
	and	al,not 10b
	wrmsr
	and	[DebugCtlMSR_settings],not 10b
	jmp	egzyduz_STATUS_SUCCESS

; ------------------------------------------------------------------------------

; dealing with exceptions:
; interrupt_descriptor_table:
; bits
; 79-16		Interrupt Descriptor Table Base Address
; 15-0		Interrupt Descriptor Table Limit = 256*16-1 = limit of IDT (size minus one)
; exception_gates:
; bit(s)
; 127-96	Reserved, IGN
; 95-64		bits 63.32. of the Target Offset
; 65-48		bits 31.16. of the Target Offset
; 47		P
; 46-45		DPL
; 44		0
; 43-40		Type
; 39-35		Reserved, IGN
; 34-32		IST
; 31-16		Target Selector
; 15-0		bits 15.0. of the Target Offset
; sample of exception01:
;	dw	exception01 and 0FFFFh,CODE64_SELECTOR
;	dw	8E01h,exception01 shr 16
;	dd	0,0

; we don't load new idst (lidt) to avoid the necessity to do it for every CPU on multiCPU system
; we rather hook vectors in memory, because memory is identical for every CPU

align 16
DispatchWrite_proc08:
HookExceptions:

; check whether exception vectors haven't already been hooked
;	cmp	[exceptions_hooked],0
;	jnz	HookExceptions_done	; we have already hooked it, don't do it again...
; eax=0 now
bts dword [exceptions_hooked],eax
jc HookExceptions_done

;	pushf
;	push	rbx			; we needn't to protect RBX because it is updated to STATUS_SUCCESS after HookExceptions_done

; Make sure that this procedure executes only on one CPU.
; Disabling interrupts works fine to avoid context switches.
; We mustn't forget to reenable interrupts at the end !!!
	cli

; OK, hook it
	sidt	[rsp-10h]
	mov	rdx,[rsp-10h+2]		; address of interrupt vectors

; temporarily disable possible protection mechanism
	mov	rax,cr0
	push	rax
;	and	eax,not (1 shl 16)	; disable CR0.WP (Write Protect Bit)
;					; Read-only pages are protected from supervisor-level writes when the WP bit is set to 1. When WP is cleared to 0, supervisor software can write into read-only pages.
	btr	eax,16			; smaller than the above and eax,...

	mov	cr0,rax

; replace addresses for int00-int1F
; we will do it from the back

size_of_exception_handler	=	20h

if size_of_exception_handler <> 20h
change indexes in following instructions:
end if

;	mov	ecx,(number_of_exceptions-1)*4
	push	(number_of_exceptions-1)*4
	pop	rcx
	lea	r9,[exception_pointers]
	lea	r8,[exception00]

AMDPad16
HL0:
; get original address
	mov	rbx,[rdx+rcx*4+4]	; get bits 95-48 of gate into bits 63-16 of RBX
	mov	bx,[rdx+rcx*4+0]	; get bits 15-0 of gate into bits 15-0 of RBX
					; note, this doesn't clear upper part of RBX in contrast of operation with 32-bit EDX !!!
; prepare new address
	lea	rax,[r8+rcx*8]		; point to exceptionXX handler

	mov	[r9+rcx*2],rbx		; save original address
					; we will use it to jump to the original address

	mov	[rdx+rcx*4+0],ax
;	shr	rax,16
;	mov	[rdx+rcx*4+6],ax
;	shr	rax,16
;	mov	[rdx+rcx*4+8],eax
	mov	ax,[rdx+rcx*4+4]
	mov	[rdx+rcx*4+4],rax
; another trick is shr rax,16 \ mov [rdx+rcx*4+6],rax if we don't worry about bits 111-96 (bits 111-96 should be zero, so this method shouldn't seem to lead to any troubble...)

; we don't change Target Selector (mov ax,cs \ mov [rdx+rcx*4+2],ax) - we assume that kernel and our driver have the same CS (code selector)
; it they differs, we can't use jmp qword [exception_pointers + 00h*8] etc. either...

	sub	ecx,4
	jns	HL0

; restore possible protection mechanism
	pop	rax
	mov	cr0,rax

; we mustn't forget to enable interrupts, else the whole OS hangs...
	sti				; we may popf for restoring rflags.IF (interrupt flag bit of rflags)

;	pop	rbx
;	popf
;	mov	[exceptions_hooked],1	; unnecessary, we have already done it by bts dword [exceptions_hooked],eax

HookExceptions_done:
	jmp	egzyduz_STATUS_SUCCESS


align 16
DispatchWrite_proc09:
	call	UnhookExceptions
	jmp	egzyduz_STATUS_SUCCESS


align 16
UnhookExceptions:
; check whether exception vectors have already been hooked and weren't unhooked
	btr	dword [exceptions_hooked],0
	jnc	UnhookExceptions_end	; they haven't been hooked yet or they were already been unhooked, we don't have anything to restore...

;	pushf
; Make sure that this procedure executes only on one CPU.
; Disabling interrupts works fine to avoid context swithches.
; We mustn't forget to reenable interrupts at the end !!!
	cli

; OK, unhook it
	sidt	[rsp-10h]
	mov	rdx,[rsp-10h+2]		; address of interrupt vectors

; temporarily disable possible protection mechanism
	mov	rax,cr0
	push	rax
	and	eax,not (1 shl 16)	; disable CR0.WP (Write Protect Bit)
					; Read-only pages are protected from supervisor-level writes when the WP bit is set to 1. When WP is cleared to 0, supervisor software can write into read-only pages.
	mov	cr0,rax

; restore addresses for int00-int1F
; we will do it from the back
	mov	ecx,(number_of_exceptions-1)*2	; *2 because we use sub ecx,2
	lea	r9,[exception_pointers]

AMDPad16
UL0:	mov	rax,[r9+rcx*4]		; get original address
	mov	[rdx+rcx*8+0],ax	; restore bits 15-0 of gate from bits 15-0 of RAX
	mov	ax,[rdx+rcx*8+4]
	mov	[rdx+rcx*8+4],rax	; restrore bits 95-48 of gate from bits 63-16 of RAX
	sub	ecx,2
	jns	UL0

; restore possible protection mechanism
	pop	rax
	mov	cr0,rax

; we mustn't forget to enable interrupts, else the whole OS hangs...
	sti				; we may popf for restoring rflags.IF (interrupt flag bit of rflags)

;	popf

UnhookExceptions_end:
	ret


macro	save_branches
{
	push	rax rcx rdx rbx rbp

	mov	ecx,LastBranchFromIP
	rdmsr

	push	rdx
	pop	rbx

	xchg	ebp,eax

if (LastBranchToIP - LastBranchFromIP) = 1
	inc	ecx
else
	mov	ecx,LastBranchToIP
end if
	rdmsr

; the rest can be saved after call (LastBranchFromIP, LastBranchToIP are destroyed after call changes RIP)
	call	save_rest
}

align 16
save_rest:
	push	rdi
	lea	rdi,[LastBranchesAndExceptions]

	mov	[rdi + LastBranchFromIPHi - LastBranchesAndExceptions],ebx
	mov	[rdi + LastBranchFromIPLo - LastBranchesAndExceptions],ebp

	mov	[rdi + LastBranchToIPHi - LastBranchesAndExceptions],edx
	mov	[rdi + LastBranchToIPLo - LastBranchesAndExceptions],eax

if (LastExceptionFromIP - LastBranchToIP) = 1
	inc	ecx
else
	mov	ecx,LastExceptionFromIP
end if
	rdmsr
	mov	[rdi + LastExceptionFromIPHi - LastBranchesAndExceptions],edx
	mov	[rdi + LastExceptionFromIPLo - LastBranchesAndExceptions],eax

if (LastExceptionToIP - LastExceptionFromIP) = 1
	inc	ecx
else
	mov	ecx,LastExceptionToIP
end if
	rdmsr

	mov	[rdi + LastExceptionToIPHi - LastBranchesAndExceptions],edx
	mov	[rdi + LastExceptionToIPLo - LastBranchesAndExceptions],eax

	pop	rdi

	pop	rax			; get return address

	pop	rbp rbx rdx rcx

	xchg	[rsp],rax		; restore rax and put return address there

	ret


; Each exception handler must have the same size. Align size_of_exception_handler does this task perfectly.
align size_of_exception_handler
exception00:
; Divide-by-Zero-Error #DE DIV, IDIV, AAM instructions
	save_branches
	jmp	qword [exception_pointers + 00h*8]
if ($ - exception00) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception01:
; Debug #DB Instruction accesses and data accesses
; The processor automatically disables control-transfer recording when a debug exception (#DB) occurs by clearing DebugCtlMSR.LBR to 0. The contents of the control-transfer recording MSRs are not altered by the processor when the #DB occurs. Before exiting the debug-exception handler, software can set DebugCtlMSR.LBR to 1 to re-enable the recording mechanism.
; The processor automatically disables control-transfer single stepping when a debug exception (#DB) occurs by clearing DebugCtlMSR.BTF to 0. rFLAGS.TF is also cleared when a #DB exception occurs. Before exiting the debugexception handler, software must set both DebugCtlMSR.BTF and rFLAGS.TF to 1 to restart single stepping.
; We must enable these bits in DebugCtlMSR now, we won't have any opportunity to do it later because we are about passing control to OS handler very soon...
; We can save LastBranchFromIP and LastBranchToIP in subprocedure. The call instruction doesn't change them because DebugCtlMSR.LBR=0 in int01 exception handler.
	call	save_branches_int01
	call	reenable_DebugCtlMSR_cleared_by_int01
	jmp	qword [exception_pointers + 01h*8]
if ($ - exception01) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception02:
; Non-Maskable-Interrupt #NMI External NMI signal
	save_branches
	jmp	qword [exception_pointers + 02h*8]
if ($ - exception02) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception03:
; Breakpoint #BP INT3 instruction
	save_branches
	jmp	qword [exception_pointers + 03h*8]
if ($ - exception03) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception04:
; Overflow #OF INTO instruction
	save_branches
	jmp	qword [exception_pointers + 04h*8]
if ($ - exception04) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception05:
; Bound-Range #BR BOUND instruction
	save_branches
	jmp	qword [exception_pointers + 05h*8]
if ($ - exception05) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception06:
; Invalid-Opcode #UD Invalid instructions
	save_branches
	jmp	qword [exception_pointers + 06h*8]
if ($ - exception06) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception07:
; Device-Not-Available #NM x87 instructions
	save_branches
	jmp	qword [exception_pointers + 07h*8]
if ($ - exception07) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception08:
; Double-Fault #DF Exception during the handling of another exception or interrupt
	save_branches
	jmp	qword [exception_pointers + 08h*8]
if ($ - exception08) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception09:
; Coprocessor-Segment-Overrun  Unsupported (Reserved)
	save_branches
	jmp	qword [exception_pointers + 09h*8]
if ($ - exception09) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception0A:
; Invalid-TSS #TS Task-state segment access and task switch
	save_branches
	jmp	qword [exception_pointers + 0Ah*8]
if ($ - exception0A) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception0B:
; Segment-Not-Present #NP Segment register loads
	save_branches
	jmp	qword [exception_pointers + 0Bh*8]
if ($ - exception0B) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception0C:
; Stack #SS SS register loads and stack references
	save_branches
	jmp	qword [exception_pointers + 0Ch*8]
if ($ - exception0C) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception0D:
; General-Protection #GP Memory accesses and protection checks
	save_branches
	jmp	qword [exception_pointers + 0Dh*8]
if ($ - exception0D) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception0E:
; Page-Fault #PF Memory accesses when paging enabled
;
; why the hell this destroys branches
; win uses this exception for some purpose ???
; this must be commented not to overwrite branches with useless addresses from kernel
; Page-Fault Memory accesses when paging enabled
;	call	save_branches_from_user_mode_exception0E
; Present (P) Bit. Bit 0. This bit indicates whether the pagetranslation table or physical page is loaded in physical memory.
; When the P bit is cleared to 0, the table or physical page is not loaded in physical memory. When the P bit is set to 1, the table or physical page is loaded in physical memory.
; Software clears this bit to 0 to indicate a page table or physical page is not loaded in physical memory. A page-fault exception (#PF) occurs if an attempt is made to access a table or page when the P bit is 0. System software is responsible for loading the missing table or page into memory and setting the P bit to 1.
; When the P bit is 0, indicating a not-present page, all remaining bits in the page data-structure entry are available to software.
; we assume win64 uses 4 kB pages
; we should check
; - Page-Translation Enable (CR0.PG)
; - Physical-Address Extensions (CR4.PAE)
; - Page-Size Extensions (CR4.PSE)
; - Long-Mode Active (EFER.LMA)
;
; push rax rcx rdx
; mov rcx,not 111111111b
; mov rax,cr3 ; pointer to PML4
; and rax,not (11000b) ; mask out CR3.PCD CR3.PWT bits - only to be sure...
; test rax,1
; jz _swapping_mem
; and rax,rcx 
;
; mov rdx,cr2 ; linear (virtual) address caused the exception
; get 9 bits of PML4 offset
; shr rdx,39
; and edx,ecx
; mov rax,[rax+rdx*8]
; test rax,1
; jz _swapping_mem
; and rax,rcx
;
; mov rdx,cr2 ; linear (virtual) address caused the exception
; get 9 bits of PDP offset
; shr rdx,30
; and edx,ecx
; mov rax,[rax+rdx*8]
; test rax,1
; jz _swapping_mem
; and rax,rcx
;
; mov rdx,cr2 ; linear (virtual) address caused the exception
; get 9 bits of PD offset
; shr rdx,21
; and edx,ecx
; mov rax,[rax+rdx*8]
; test rax,1
; jz _swapping_mem
; and rax,rcx
;
; mov rdx,cr2 ; linear (virtual) address caused the exception
; get 9 bits of PT offset
; shr rdx,12
; and edx,ecx
; mov rax,[rax+rdx*8]
; test rax,1
; jz _swapping_mem
; pop rdx rcx rax
;
;_swapping_mem:
; pop rdx rcx rax
; the easiest is testing CS pushed in stack, e.g. CS=8 is kernel, CS=8+11b is user mode
; Error Code Returned. The page-fault error code is pushed onto the page-fault exception-handler stack.
; [rsp+8*0] error code
; [rsp+8*1] RIP
; [rsp+8*2] CS
; [rsp+8*3] RSP
; [rsp+8*4] SS

; test byte [rsp+10h],11b

; fuck it off...
	save_branches
	jmp	qword [exception_pointers + 0Eh*8]
if ($ - exception0E) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception0F:
; Reserved
	jmp	qword [exception_pointers + 0Fh*8]
if ($ - exception0F) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception10:
; x87 Floating-Point Exception-Pending #MF x87 floating-point instructions
	save_branches
	jmp	qword [exception_pointers + 10h*8]
if ($ - exception10) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception11:
; Alignment-Check #AC Misaligned memory accesses
	save_branches
	jmp	qword [exception_pointers + 11h*8]
if ($ - exception11) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception12:
; Machine-Check #MC Model specific
	save_branches
	jmp	qword [exception_pointers + 12h*8]
if ($ - exception12) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception13:
; SIMD Floating-Point #XF 128-bit media floating-point instructions
	save_branches
	jmp	qword [exception_pointers + 13h*8]
if ($ - exception13) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception14:
; Reserved
	jmp	qword [exception_pointers + 14h*8]
if ($ - exception14) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception15:
; Reserved
	jmp	qword [exception_pointers + 15h*8]
if ($ - exception15) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception16:
; Reserved
	jmp	qword [exception_pointers + 16h*8]
if ($ - exception16) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception17:
; Reserved
	jmp	qword [exception_pointers + 17h*8]
if ($ - exception17) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception18:
; Reserved
	jmp	qword [exception_pointers + 18h*8]
if ($ - exception18) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception19:
; Reserved
	jmp	qword [exception_pointers + 19h*8]
if ($ - exception19) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception1A:
; Reserved
	jmp	qword [exception_pointers + 1Ah*8]
if ($ - exception1A) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception1B:
; Reserved
	jmp	qword [exception_pointers + 1Bh*8]
if ($ - exception1B) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception1C:
; Reserved
	jmp	qword [exception_pointers + 1Ch*8]
if ($ - exception1C) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception1D:
; Reserved
	jmp	qword [exception_pointers + 1Dh*8]
if ($ - exception1D) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception1E:
; Security Exception #SX Security-sensitive event in host
	save_branches
	jmp	qword [exception_pointers + 1Eh*8]

if ($ - exception1E) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if

align size_of_exception_handler
exception1F:
; Reserved
;	save_branches
; why the hell this destroys branches
; this must be commented not to overwrite branches with useless addresses from kernel
; we must do this to prevent OS to overwrite our branches with useless kernel branches generated at OS page-faults (OS method for loading swapped memory from disk uses Page-Faults)
; win64 uses it for something...
	jmp	qword [exception_pointers + 1Fh*8]
if ($ - exception1F) > size_of_exception_handler
adjust this handler size to fit into size_of_exception_handler
end if


align 16
save_branches_int01:
	push	rax rcx rdx

	mov	ecx,LastBranchFromIP
	rdmsr
	mov	[LastBranchFromIPHi],edx
	mov	[LastBranchFromIPLo],eax

if (LastBranchToIP - LastBranchFromIP) = 1
	inc	ecx
else
	mov	ecx,LastBranchToIP
end if
	rdmsr
	mov	[LastBranchToIPHi],edx
	mov	[LastBranchToIPLo],eax

if (LastExceptionFromIP - LastBranchToIP) = 1
	inc	ecx
else
	mov	ecx,LastExceptionFromIP
end if
	rdmsr
	mov	[LastExceptionFromIPHi],edx
	mov	[LastExceptionFromIPLo],eax

if (LastExceptionToIP - LastExceptionFromIP) = 1
	inc	ecx
else
	mov	ecx,LastExceptionToIP
end if
	rdmsr

	mov	[LastExceptionToIPHi],edx
	mov	[LastExceptionToIPLo],eax

	pop	rdx rcx rax
	ret

align 16
reenable_DebugCtlMSR_cleared_by_int01:
	push	rax rcx rdx

	mov	ecx,DebugCtlMSR
	rdmsr
	and	al,not 11b			; mask out all bits of our interest
	or	al,[DebugCtlMSR_settings]	; and put there our wishes...
	wrmsr

	pop	rdx rcx rax
	ret

;align 16
;save_branches_from_user_mode_exception0E:
;
;; rsp+8*0 -> qword return address from this procedure
;
;; Interrupt-Handler Stack with Error Code
;; rsp+8*1 -> dword Error Code
;; rsp+8*2 -> qword return RIP
;; rsp+8*3 -> word return CS
;; rsp+8*4 -> qword return rflags
;; rsp+8*5 -> qword return RSP
;; rsp+8*6 -> word return SS

;; The virtual (linear) address that caused the #PF is stored in the CR2 register.
;	push	rax
;	mov	rax,cr2
;	pop	rax

; segment selectors:
; Bits	Mnemonic	Description
; 15-3	SI		Selector Index R/W
; 2	TI		Table Indicator R/W
; 1-0	RPL		Requestor Privilege Level R/W
;	test	byte [rsp+8*1],11b		; check code segment pushed into stack
;	j	save_branches_from_user_mode_exception0E_L0
;	save_branches
;save_branches_from_user_mode_exception0E_L0:
; doesn't help, pages are loaded from swap into user mode too
;	ret


padnum16zero
cusDevice_string	du	'\Device\fdbg',0

padnum16zero
cusSymbolicLink_string	du	'\DosDevices\fdbg',0


section '.udata' readable writeable notpageable  
; Must be ultimately notpageable, holds interrupt descriptor and interrupt gates...
; Thus the section mustn't be swapped to disk !!!

align 16
exception_pointers	rb	number_of_exceptions * size_of_address
exception_pointers_end:

align 16
LastBranchesAndExceptions:
LastBranchFromIPLo	dd	?
LastBranchFromIPHi	dd	?
LastBranchToIPLo	dd	?
LastBranchToIPHi	dd	?
LastExceptionFromIPLo	dd	?
LastExceptionFromIPHi	dd	?
LastExceptionToIPLo	dd	?
LastExceptionToIPHi	dd	?

DebugCtlMSR_eax_old	dd	?
DebugCtlMSR_edx_old	dd	?

DebugCtlMSR_settings	db	?
exceptions_hooked	db	?


section '.rdata' readable writeable notpageable  

data 12  

LookupNTOSKRNL:
IoCreateDevice		dq	rva _IoCreateDevice
IoCreateSymbolicLink	dq	rva _IoCreateSymbolicLink
IoDeleteDevice		dq	rva _IoDeleteDevice
IoDeleteSymbolicLink	dq	rva _IoDeleteSymbolicLink
IofCompleteRequest	dq	rva _IofCompleteRequest
KeQueryActiveProcessors	dq	rva _KeQueryActiveProcessors
PsCreateSystemThread	dq	rva _PsCreateSystemThread
RtlInitUnicodeString	dq	rva _RtlInitUnicodeString
ZwSetInformationThread	dq	rva _ZwSetInformationThread
			dq	0

end data


section 'INIT' data import readable notpageable

			dd	rva AddressNTOSKRNL,	0, 0, rva _ntoskrnl,	rva LookupNTOSKRNL
			dd	0,			0, 0, 0,		0

AddressNTOSKRNL		dq	rva _IoCreateDevice
			dq	rva _IoCreateSymbolicLink
			dq	rva _IoDeleteDevice
			dq	rva _IoDeleteSymbolicLink
			dq	rva _IofCompleteRequest
			dq	rva _KeQueryActiveProcessors
			dq	rva _PsCreateSystemThread
			dq	rva _RtlInitUnicodeString
			dq	rva _ZwSetInformationThread
			dq	0

;AddressHAL
;			dq	0

_IoCreateDevice		db	0,0,'IoCreateDevice',0
_IoCreateSymbolicLink	db	0,0,'IoCreateSymbolicLink',0
_IoDeleteDevice		db	0,0,'IoDeleteDevice',0
_IoDeleteSymbolicLink	db	0,0,'IoDeleteSymbolicLink',0
_IofCompleteRequest	db	0,0,'IofCompleteRequest',0
_KeQueryActiveProcessors db	0,0,'KeQueryActiveProcessors',0
_PsCreateSystemThread	db	0,0,'PsCreateSystemThread',0
_RtlInitUnicodeString	db	0,0,'RtlInitUnicodeString',0
_ZwSetInformationThread	db	0,0,'ZwSetInformationThread',0

_ntoskrnl 		db	'ntoskrnl.exe',0
